public class BinTree<D> implements Dictionary<D> {
   Node<D> root;

   /**
    * finds a node with a given key in this tree 
    * @param key
    * @return null, if the root is null; 
    *    a node with key, if this exists;
    *    a leaf below which the key has to be inserted, otherwise 
    */
   Node<D> find(int key) {
      Node<D> current = this.root;
      Node<D> previous = null;
      while (current != null) {
         previous = current;
         if (key < current.key) {
            current = current.left;
         } else if (key == current.key) {
            return current;
         } else {
            current = current.right;
         }
      }
      return previous; 
   }

   public D get(int key) {
      Node<D> node = find(key);
      if (node != null && node.key == key) {
         return node.data;
      } else {
         return null;
      }
   }

   public void put(int key, D data) {
      Node<D> node = find(key);
      if (node == null) { // root is null
         this.root = new Node<D>(key,data,null,null,null);
      } else if (node.key == key) { // overwrite
         node.data = data;
      } else { // new node below node
         Node<D> newNode = new Node<D>(key,data,null,null,node);
         if (key < node.key) {
            node.left = newNode;
         } else {
            node.right = newNode;
         }
      }
   }

   /**
    * delivers the node which stores
    * the maximal key in the subtree with root "node"
    * @param node
    */
   static <D> Node<D> maximum(Node<D> node) {
      while (node.right != null) {
         node = node.right;
      }
      return node;
   }

   /**
    * removes a given non-null node from this tree
    */
   void removeNode(Node<D> node) {
      if (node.left == null && node.right == null) { // leaf
         if (node.parent == null) {
            this.root = null;
         } else {
            if (node.key < node.parent.key) {
               node.parent.left = null;
            } else {
               node.parent.right = null;
            }
         }
      } else { // non-leaf             
         if (node.left == null || node.right == null) { 
            Node<D> child = node.left==null ? node.right:node.left;
            if (node.parent == null) { // delete root
               this.root = child;
            } else if (node.key < node.parent.key) {
               node.parent.left = child;
            } else {
               node.parent.right = child;
            }
            child.parent = node.parent;
         } else { // two children
            Node<D> maxLeft = maximum(node.left);
            removeNode(maxLeft);
            node.data = maxLeft.data;
            node.key = maxLeft.key;
         }
      }      
   }

   public void remove(int key) {
      Node<D> node = find(key);
      if (node != null && node.key == key) {
         removeNode(node);
      }
   }

   public Iterator<D> iterator() {      
      return new TreeIterator<D>(this.root);
   }

}
